/*
PrintIt Copyright (c) 2005-2006, Alexander R. Pruss
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (
INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (
INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/
// Copyright (C) 2004-6 Alexander R. Pruss
//

#define ALLOW_ACCESS_TO_INTERNALS_OF_BITMAPS
#include <PalmOS.h>
#include <BmpGlue.h>
#include "PrintIt.h"

#ifndef memHeapFlagReadOnly
# define memHeapFlagReadOnly 0x0001
#endif

#define memNewChunkFlagAllowLarge 0x1000
#define dmOrphanOwnerID            3

#include "pref.c"

static UInt16 Dialog( PreferenceType* p );

static MemHandle AllocateChunk( UInt32 size, Boolean dynamic, Boolean moveable )
{
    UInt16 i;
    UInt16 n;
    UInt16 card;
    UInt16 moveableAttr;

    if ( moveable )
        moveableAttr = 0;
    else
        moveableAttr = memNewChunkFlagNonMovable;

    for ( card = 0 ; card < 1 ; card++ ) {
        n = MemNumHeaps( card );

        for ( i = 0 ; i < n ; i++ ) {
             UInt16 heapID;
             UInt16 flags;

             heapID = MemHeapID( card, i );

             flags  = MemHeapFlags( heapID );

             if ( ! ( flags & memHeapFlagReadOnly ) && !! MemHeapDynamic( heapID ) == !! dynamic ) {
                 MemHandle h;
                 h = MemChunkNew( heapID, size, dmOrphanOwnerID |
                                            memNewChunkFlagAllowLarge |
                                            moveableAttr );
                 if ( h != NULL )
                     return h;
             }
        }
    }
    return NULL;
}


static void* MemPtrNewBig( UInt32 size )
{
//    if ( size < 65000 )
//        return MemPtrNew( size );
    return AllocateChunk( size, true, false );
}

static BitmapType* GetImage( void )
{
    BitmapTypeV3* image;
    BitmapType*   source;
    Coord width;
    Coord height;
    UInt8 depth;
    UInt16 rowBytes;
    Boolean   hiRes;
    Err       err;
    UInt32    version;
    Boolean   copyManually;
    WinHandle oldWin;
    UInt16    oldCoords;

    err = FtrGet( sysFtrCreator, sysFtrNumWinVersion, &version );
    hiRes = ( err == errNone && 4 <= version );

    oldWin    = WinSetDrawWindow( WinGetDisplayWindow() );
    oldCoords = hiRes ? WinSetCoordinateSystem( kCoordinatesNative ) : 72;

    source    = WinGetBitmap( WinGetDisplayWindow() );

    WinGetDisplayExtent( &width, &height );

    depth = BmpGlueGetBitDepth( source );

    copyManually = ! ( hiRes && 1 < depth );
    depth        = 4;

    width = ( width + 3 ) / 4 * 4;

    if ( depth <= 8 ) {
        UInt8     pixelsPerByte;

        pixelsPerByte = 8 / depth;

        rowBytes = ( width + pixelsPerByte - 1 ) / pixelsPerByte;
    }
    else {
        rowBytes = width * ( UInt32 )depth / 8;
    }

    image = MemPtrNewBig( sizeof( BitmapTypeV3 ) + ( ( UInt32 )rowBytes * height ) );

    if ( image != NULL ) {
        WinHandle newWin;

        MemZero( image, sizeof( BitmapTypeV3 ) );
        image->width = width;
        image->height = height;
        image->rowBytes = rowBytes;
        image->pixelSize = depth;
        if ( 16 <= depth )
            image->pixelFormat = pixelFormat565;
        else
            image->pixelFormat = pixelFormatIndexed;

        if ( hiRes ) {
            image->version   = 3;
            image->size      = sizeof( BitmapTypeV3 );
            image->density     = BmpGetDensity( source );
        }
        else {
            image->version   = 2;
        }

        /* does not support hi-res now */
        if ( copyManually ) {
            WinHandle oldWin;
            Coord     x;
            Coord     y;
            UInt8*    bits;

            bits   = BmpGetBits( ( BitmapType* )image );

            oldWin = WinSetDrawWindow( WinGetDisplayWindow() );

            for ( y = 0 ; y < height ; y++ ) {
                 UInt8* out = bits + ( UInt32 )rowBytes * y;

                 for ( x = 0 ; x < width ; x += 2 ) {
                      *out++ = ( GetPixel( x, y, 15 ) << 4 ) |
                               ( GetPixel( x + 1, y, 15 ) );
                 }
            }

            WinSetDrawWindow( oldWin );
        }
        else {
            newWin = WinCreateBitmapWindow( ( BitmapType* )image, &err );

            if ( newWin != NULL ) {
                RectangleType rect;

                rect.topLeft.x = 0;
                rect.topLeft.y = 0;
                rect.extent.x  = width;
                rect.extent.y  = height;

                WinCopyRectangle( WinGetDisplayWindow(), newWin, &rect, 0, 0, winPaint );

                WinDeleteWindow( newWin, false );
            }
            else {
                MemPtrFree( image );
                image = NULL;
            }
        }

    }

    if ( oldWin != NULL )
        WinSetDrawWindow( oldWin );
    if ( hiRes )
        WinSetCoordinateSystem( oldCoords );

    return ( BitmapType* )image;
}



void Go( void )
{
    UInt16 button;
    BitmapType* bmp;
    PrintScreenData data;
    
    data.timeStamp = TimGetSeconds();

    bmp = GetImage();

    if ( bmp == NULL )
        return;

    do {
        PreferenceType* p;

        button = btnQuit;

        p = MemPtrNew( sizeof( PreferenceType ) );

        if ( p != NULL ) {
            LoadPrefs0( p );

            button = Dialog( p );

            switch( button ) {
                case btnPrint:
                case btnTest: {
                    Boolean success;

                    NetOpen();
                    p->jobNumber =  p->jobNumber + 1;
                    SavePrefs0( p );

                    if ( btnTest == button ) {
                        success = Print( &( p->history[ 0 ] ), PrinterTestDriver, NULL, "PS/PCL Test", p->jobNumber );
                    }
                    else {
                        data.bitmap = bmp;
                        data.big    = p->doubleSize;
                        data.force600DPI = ( p->history[ 0 ].resolution == 600 ) && p->history[ 0 ].usePJL;

                        success = Print( &( p->history[ 0 ] ), PrintScreenDriver, &data, "PrintIt Screen Print", p->jobNumber );
                    }

                    NetClose();

                    if ( ! success )
                        FrmAlert( altError );

                    break;
                }
            }

            MemPtrFree( p );
        }
    } while ( button == btnTest );

    MemPtrFree( bmp );
}


FormType* GetMyForm( void )
{
    FormType* f;

    FtrGet( appCreator, FTR_FORM, ( UInt32* )&f );

    return f;
}



UInt16 GetObjectIndex( UInt16 id )
{
    return FrmGetObjectIndex( GetMyForm(), id );
}


void* GetObjectPtr( UInt16 id )
{
    return FrmGetObjectPtr( GetMyForm(), GetObjectIndex( id ) );
}


Int16 GetSelection( UInt16 lstID )
{
    return LstGetSelection( GetObjectPtr( lstID ) );
}


void Show( UInt16 id, Boolean state )
{
    UInt16 index = GetObjectIndex( id );
    FormType* currentForm = GetMyForm();

    if ( state )
        FrmShowObject( currentForm, index );
    else
        FrmHideObject( currentForm, index );
    if ( FrmGetObjectType( currentForm, index ) == frmControlObj )
        CtlSetUsable( GetObjectPtr( id ), state);
}


void SetList( UInt16 trgID, UInt16 lstID, Int16 sel )
{
    LstSetSelection( GetObjectPtr( lstID ),
        sel );
    CtlSetLabel( GetObjectPtr( trgID ),
        LstGetSelectionText( GetObjectPtr( lstID ),
        sel ) );
}

void SetValue( UInt16 id, UInt16 value )
{
    CtlSetValue( GetObjectPtr( id ), value );
}


UInt16 GetValue( UInt16 id )
{
    return CtlGetValue( GetObjectPtr( id ) );
}


void GetField( Char* s, UInt16 id )
{
    FieldType* f;
    Char*      text;

    f = GetObjectPtr( id );

    text = FldGetTextPtr( f );

    if ( text == NULL )
        *s = 0;
    else
        StrCopy( s, text );
}


void SetField( UInt16 id, Char* s )
{
    FieldType* f;

    f = GetObjectPtr( id );

    FldDelete( f, 0, 0xFFFF );
    FldInsert( f, s, StrLen( s ) );
}



void SelectAllText( UInt16 id )
{
    FldSetSelection( GetObjectPtr( id ), 0, 0xFFFF );
}


static void ShowPrinter( PrinterDescriptor* printer )
{
    Show( lblQueue, printer->protocol == PRINT_LPD );
    Show( fldQueue, printer->protocol == PRINT_LPD );
    if ( printer->protocol == PRINT_LPD )
        SetField( fldQueue, printer->queue );
    SetField( fldAddress, printer->address );
    SetList( trgProtocol, lstProtocol, printer->protocol );
    SetList( trgDriver, lstDriver, printer->driver );
    SetValue( chkUsePJL, printer->usePJL );
    if ( printer->driver == DRIVER_PCL && ! GetValue( chkDouble ) &&
         GetValue( chkUsePJL ) ) {
        SetValue( chkForce600DPI, printer->resolution == 600 );
        Show( chkForce600DPI, true );
    }
    else {
        Show( chkForce600DPI, false );
    }
}



static void PopUpHistory( PreferenceType* p, PrinterDescriptor* printer )
{
    Char** names;
    UInt16 i;
    UInt16 numNames;
    Int16  sel;
    ListType* list;

    if ( ! p->history[ 0 ].address[ 0 ] )
        return;

    names = MemPtrNew( sizeof( Char* ) * NUM_HISTORY_ENTRIES );

    if ( names == NULL )
        return;

    for ( i = 0 ; i < NUM_HISTORY_ENTRIES && p->history[ i ].address[ 0 ] ; i++ ) {
         UInt16 length;

         length = StrLen( p->history[ i ].address );

         if ( p->history[ i ].protocol == PRINT_LPD )
             length += 1 + StrLen( p->history[ i ].queue );

         names[ i ] = MemPtrNew( length + 1 );

         if ( names[ i ] == NULL ) {
             UInt16 j;

             for ( j = 0 ; j < i ; j++ )
                  MemPtrFree( names[ j ] );

             MemPtrFree( names );
         }

         switch ( p->history[ i ].protocol ) {
             case PRINT_RAW:
                 StrCopy( names[ i ], p->history[ i ].address );
                 break;
             case PRINT_LPD: {
                 UInt16 qlen;

                 qlen = StrLen( p->history[ i ].queue );
                 StrCopy( names[ i ], p->history[ i ].queue );
                 names[ i ][ qlen ] = '@';
                 StrCopy( names[ i ] + qlen + 1, p->history[ i ].address );
             }
         }
    }


    numNames = i;
    
    list = GetObjectPtr( lstHistory );

    LstSetHeight( list, numNames );

    LstSetListChoices( list, names, numNames );
    LstSetSelection( list, 0 );

    sel = LstPopupList( list );

    for ( i = 0 ; i < numNames ; i++ )
         MemPtrFree( names[ i ] );
    MemPtrFree( names );

    if ( 0 <= sel ) {
        *printer = p->history[ sel ];
        ShowPrinter( printer );
    }
}



static void GetPrinterFromFields( PrinterDescriptor* printer )
{
    GetField( printer->address, fldAddress );
    if ( printer->protocol == PRINT_LPD )
        GetField( printer->queue, fldQueue);
    printer->usePJL = !! GetValue( chkUsePJL );
    if ( printer->usePJL ) {
        printer->resolution = GetValue( chkForce600DPI ) ? 600 : 300;
    }
    else {
        printer->resolution = true;
    }
}



static Boolean DialogHandler( EventType* event )
{
    Boolean handled;
    PreferenceType* p;
    PrinterDescriptor* printer;
    UInt32             flags;

    handled = false;

    FtrGet( appCreator, FTR_PREFERENCE, ( UInt32* )&p );
    FtrGet( appCreator, FTR_PRINTER, ( UInt32* )&printer );
    FtrGet( appCreator, FTR_FLAGS, &flags );

    switch ( event->eType ) {
        case popSelectEvent: {
            Int16 item;

            item = event->data.popSelect.selection;

            switch ( event->data.popSelect.controlID ) {
                case trgProtocol:
                    printer->protocol = item;
                    GetPrinterFromFields( printer );
                    ShowPrinter( printer );
//                    handled = true;
                    break;
                case trgDriver:
                    printer->driver = item;
                    SetList( trgDriver, lstDriver, printer->driver );
                    GetPrinterFromFields( printer );
                    ShowPrinter( printer );
//                    handled = true;
                    break;
            }
        }

        case frmCloseEvent: {
            handled = true;
            break;
        }

        case ctlSelectEvent: {
            switch( event->data.ctlSelect.controlID ) {
                case chkUsePJL:
                    printer->usePJL = !! GetValue( chkUsePJL );
                    GetPrinterFromFields( printer );
                    ShowPrinter( printer );
                    handled = true;
                    break;
                case chkDouble:
                    p->doubleSize = !! GetValue( chkDouble );
                    SavePrefs0( p );
                    GetPrinterFromFields( printer );
                    ShowPrinter( printer );
                    handled = true;
                    break;
                case trgHistory:
                    PopUpHistory( p, printer );
                    handled = true;
                    break;
                case btnClear:
                    printer->address[ 0 ] = 0;
                    printer->queue[ 0 ] = 0;
                    SetField( fldAddress, printer->address );
                    SetField( fldQueue, printer->queue );
                    printer->protocol = PRINT_RAW;
                    printer->driver   = DRIVER_PCL;
                    p->doubleSize = 0;
                    SetValue( chkDouble, 0 );
                    printer->usePJL = false;
                    printer->resolution = 0;
                    SavePrefs0( p );
                    ShowPrinter( printer );
                    handled = true;
                    break;
                case btnTest:
                case btnPrint:
                    GetPrinterFromFields( printer );
                    if ( ! printer->address[ 0 ] ||
                         ( ( printer->protocol == PRINT_LPD ) && ! printer->queue[ 0 ] ) ) {
                        FrmAlert( ( printer->protocol == PRINT_LPD ) ? altNeedInfo_LPD : altNeedInfo_RAW );
                        handled = true;
                    }
                    break;
            }
        }

        default:
            break;
    }

    return handled;
}



static UInt16 Dialog( PreferenceType* p )
{
    FormType* f;
    UInt16    button;
    PrinterDescriptor* printer;
//    EventType          e;

    f = FrmInitForm( frmPopup );

    if ( f == NULL )
        return btnQuit;

    printer = MemPtrNew( sizeof( PrinterDescriptor ) );

    if ( printer == NULL ) {
        FrmDeleteForm( f );
        return btnQuit;
    }

    FtrSet( appCreator, FTR_PREFERENCE, ( UInt32 )p );
    FtrSet( appCreator, FTR_PRINTER, ( UInt32 )printer );
    FtrSet( appCreator, FTR_FORM,    ( UInt32 )f );
    FtrSet( appCreator, FTR_FLAGS, 0 );

    *printer = p->history[ 0 ];

    ShowPrinter( printer );
    SetValue( chkDouble, p->doubleSize );
    Show( trgHistory, !! printer->address[ 0 ] );

    FrmSetEventHandler( f, DialogHandler );
    button = FrmDoDialog( f );

#if 0
    FrmSetActiveForm( f );
    FrmDrawForm( f );

    do {
        EvtGetEvent( &e, evtWaitForever );
    } while ( SysHandleEvent( &e ) || DialogHandler( &e ) ||
              FrmHandleEvent( f, &e ) || e.eType != ctlSelectEvent );

    FrmEraseForm( f );

    button = e.data.ctlSelect.controlID;
#endif

    if ( button != btnQuit ) {
        UInt16 i;

        for ( i = 0 ; i < NUM_HISTORY_ENTRIES - 1 ; i++ ) {
             if ( ! StrCompare( printer->address, p->history[ i ].address ) &&
                  ! ( printer->protocol == PRINT_LPD && p->history[ i ].protocol == PRINT_LPD &&
                      StrCompare( printer->queue, p->history[ i ].queue ) ) ) {
                 break;
             }
        }

        if ( 0 < i ) {
            MemMove( p->history + 1, p->history, sizeof( PrinterDescriptor ) * i );
        }

        p->history[ 0 ] = *printer;
    }

    FrmDeleteForm( f );

    MemPtrFree( printer );
    SavePrefs0( p );

    return button;
}




UInt32 PilotMain( UInt16 command, void* cmdPBP, UInt16 flags )
{
    switch ( command ) {
        case DO_IMAGE:
            Go();
            break;
    }
    return 0;
}


