/*
PrintIt Copyright (c) 2005-2006, Alexander R. Pruss
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (
INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (
INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/
/* $Id: PrintProtocol.c,v 1.6 2006/07/04 11:47:59 arpruss Exp $ */
#include <PalmOS.h>
#include <NetMgr.h>
#include <VFSMgr.h>
#include <Unix/sys_socket.h>
#include <System/DLServer.h>

#include "PrintProtocol.h"

#define TIMEOUT              60

#define LINGER_TIME          60

#define MAX_HOTSYNC_NAME     64
#define HEADER_SIZE 1024

NetSocketRef controlSocket = -1;
Boolean             netOpen = false;
UInt32              dataLength;
UInt16 errno;

static Int32 oldTimeout = -1;

static Char hotSyncName[ dlkUserNameBufSize ];

static Char* printerLanguages[ NUM_DRIVERS ] =
{
     "PCL",
     "POSTSCRIPT"
};

static UInt16 ports[ NUM_PROTOCOLS ] =
{
     9100, /* RAW */
     515   /* LPD */
};

void Linger( NetSocketRef socket, Boolean state, UInt16 time )
{
    struct linger l;

    l.l_onoff = state;
    l.l_linger  = time;

    setsockopt( socket, SOL_SOCKET, SO_LINGER, &l, sizeof( l ) );
}


static void NetCloseGeneric( Boolean immediate )
{
    if ( 0 <= oldTimeout ) {
        AppNetTimeout = oldTimeout;
        oldTimeout = -1;
    }

//    SysTaskDelay( SysTicksPerSecond() * 2 );

    if ( netOpen ) {
        NetLibClose( AppNetRefnum, false ); //non-immediate close
        netOpen = false;
    }
}



void NetClose( void )
{
    NetCloseGeneric( false );
}



void NetCloseImmediate( void )
{
    NetCloseGeneric( true );
}


Boolean NetOpen( void )
{
//    UInt32 ifCreator;
//    UInt16 ifInstance;
    Err    err;
    UInt16 returnValue;

    NetClose();

    errno = 0;

    err = SysLibFind( "Net.lib", &AppNetRefnum );

    if ( err != errNone ) {
        return false;
    }

#if 0
    err = NetLibIFGet( AppNetRefnum, 0, &ifCreator, &ifInstance );

    if ( err != errNone ) {
        return false;
    }
#endif

    err = NetLibOpen( AppNetRefnum, &returnValue );

    if ( err != netErrAlreadyOpen && err != errNone ) {
        return false;
    }

    oldTimeout = AppNetTimeout;
    
    AppNetTimeout = TIMEOUT * ( UInt32 )SysTicksPerSecond();

    netOpen    = true;

    return true;
}



Boolean SendString( Char* s )
{
    UInt16 len;

    len = StrLen( s );

    return NetUWriteN( controlSocket, s, len ) == len;
}


Boolean SendNull( void )
{
    Char zero;
    zero = 0;

    return 1 == NetUWriteN( controlSocket, &zero, 1 );
}


void ClosePrint( void )
{

    if ( controlSocket < 0 )
        return;

    close( controlSocket );
    controlSocket = -1;
}


Boolean GetAck_LPD( void )
{
    Char c;
    Boolean result;

    result = ( NetUReadN( controlSocket, &c, 1 ) == 1 && c == 0 );

    return result;
}



/* This initializes communication with lpd */
Boolean SendHeaders_LPD( UInt32 dataSize, Char* lpName, Char* jobTitle, UInt16 jobNumber )
{
    Char* header;
    Char  jobName[ 4 + 7 + dlkUserNameBufSize + 1 ];
    Char  commandLine[ 4 + 7 + dlkUserNameBufSize + 1 + 30 ];
    UInt16 lpNameLength;

    jobName[ 0 ] = 'A';
    jobName[ 1 ] = ( jobNumber / 100 ) % 10 + '0';
    jobName[ 2 ] = ( jobNumber / 10 ) % 10 + '0';
    jobName[ 3 ] = ( jobNumber ) % 10 + '0';
    StrCopy( jobName + 4, "PalmOS_" );
    StrCat( jobName, hotSyncName );

    header = MemPtrNew( HEADER_SIZE );

    if ( header == NULL )
        return false;

    header[ 0 ] = '\2';
    lpNameLength = StrLen( lpName );
    StrCopy( header + 1, lpName );
    header[ 1 + lpNameLength ] = '\n';
    header[ 2 + lpNameLength ] = 0;
    
    if ( ! SendString( header ) || ! GetAck_LPD() ) {
        MemPtrFree( header );
        return false;
    }

    StrPrintF( header, "ldf%s\n"
                       "Udf%s\n"
                       "P%s\n"
                       "HPalmOS_%s\n"
                       "J%s\n"
                       "N(screen)\n"
                       "T%s\n",
               jobName, jobName, hotSyncName, hotSyncName, jobName, jobTitle );

    StrPrintF( commandLine, "\2%u cf%s\n", StrLen( header ), jobName );
    SendString( commandLine );

    if ( ! GetAck_LPD() ) {
        MemPtrFree( header );
        return false;
    }

    SendString( header );
    SendNull();

    if ( ! GetAck_LPD() ) {
        MemPtrFree( header );
        return false;
    }

    StrPrintF( commandLine, "\3%lu df%s\n", dataSize, jobName );
    SendString( commandLine );

    return GetAck_LPD();
}


void GetHotSyncName( void )
{
    Char* p;

    DlkGetSyncInfo( NULL, NULL, NULL, hotSyncName, NULL, NULL );

    if ( hotSyncName[ 0 ] == 0 ) {
        StrCopy( hotSyncName, "PalmUser" );
        return;
    }

    hotSyncName[ dlkUserNameBufSize - 1 ] = 0;

    p = hotSyncName;

    while ( *p ) {
        if ( ! ( ( '0' <= *p && *p <= '9' ) || ( 'a' <= *p && *p <= 'z' ) ||
                 ( 'A' <= *p && *p <= 'Z' ) ) )
            *p = '_';

        p++;
    }

}


Boolean MeasureData( UInt32 ref, void* data, Int32 dataSize )
{
    if ( dataSize < 0 )
        dataSize = StrLen( data );

    dataLength += dataSize;

    return true;
}



Boolean SendProtocolData( UInt32 ref, void* data, Int32 dataSize )
{
    if ( dataSize < 0 )
        dataSize = StrLen( data );

    dataLength += dataSize;

    return dataSize == NetUWriteN( controlSocket, data, dataSize );
}



Boolean PJLSettings( PrinterDescriptor* printerP, ProtocolSendFunction SendFunction,
    UInt32 sendFunctionRef, Boolean dataMatters )
{
    Char header[ 80 ];

    if ( ! printerP->usePJL )
        return true;

    if ( ! SendFunction( sendFunctionRef, "\x1B%-12345X@PJL JOB\r\n", -1 ) ) {
        return false;
    }

    if ( printerP->resolution != 0 ) {
        StrPrintF( header, "@PJL SET RESOLUTION=%u\r\n", printerP->resolution );
        if ( ! SendFunction( sendFunctionRef, header, -1 ) )
            return false;
    }

    StrPrintF( header, "@PJL ENTER LANGUAGE=%s\r\n", printerLanguages[ printerP->driver ] );

    return SendFunction( sendFunctionRef, header, -1 );
}



Boolean PJLTrailer( PrinterDescriptor* printerP, ProtocolSendFunction SendFunction,
    UInt32 sendFunctionRef, Boolean dataMatters )
{
    if ( ! printerP->usePJL )
        return true;

    return SendFunction( sendFunctionRef, "\x1B%-12345X@PJL EOJ\r\n\x1B%-12345X", -1 );
}


Boolean SendJob( PrinterDescriptor* printerP, PrintDriverFunction DriverFunction, 
    void* data, ProtocolSendFunction SendFunction,
    UInt32 sendFunctionRef, Boolean dataMatters )
{
    return PJLSettings( printerP, SendFunction, sendFunctionRef, dataMatters ) &&
           DriverFunction( printerP->driver, data, SendFunction, sendFunctionRef, dataMatters ) &&
           PJLTrailer( printerP, SendFunction, sendFunctionRef, dataMatters );
}


Boolean Print( PrinterDescriptor* printerP, PrintDriverFunction DriverFunction, void* data,
    Char* jobTitle, UInt32 jobNumber )
{
    Boolean   result;

    result = false;

    if ( ! netOpen )
        return false;

    GetHotSyncName();

    controlSocket = -1;

    if ( printerP->protocol == PRINT_LPD ) {

        dataLength    = 0;

        if ( SendJob( printerP, DriverFunction, data, MeasureData, 0, false ) )
            goto CLOSE;
    }

    controlSocket = NetUTCPOpen( printerP->address, NULL, ports[ printerP->protocol ] );

    if ( controlSocket < 0 )
        goto CLOSE;

    Linger( controlSocket, true, LINGER_TIME * SysTicksPerSecond() );

    if ( printerP->protocol == PRINT_LPD &&
         ! SendHeaders_LPD( dataLength, printerP->queue, jobTitle, jobNumber % 1000 ) ) {
        goto CLOSE;
    }

    result = SendJob( printerP, DriverFunction, data, SendProtocolData, 0, true );

    if ( result && printerP->protocol == PRINT_LPD )
        result = SendNull();

CLOSE:
    if ( 0 <= controlSocket )
        ClosePrint();

    return result;
}




