/*
PrintIt Copyright (c) 2005-2006, Alexander R. Pruss
All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

3. The name of the author may not be used to endorse or promote products derived
from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (
INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (
INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/
#ifndef _PRINTPROTOCOL_H
#define _PRINTPROTOCOL_H

#include "PrintProtocolLimits.h"

typedef enum {
    DRIVER_PCL = 0,
    DRIVER_PS
} DriverType;

#define NUM_DRIVERS 2

typedef enum {
    PRINT_RAW = 0,
    PRINT_LPD
} ProtocolType;

#define NUM_PROTOCOLS 2

typedef struct {
    Char address[ MAX_SERVER_SIZE + 1 ];
    Char queue[ MAX_QUEUE_SIZE + 1 ];
    DriverType driver;
    ProtocolType protocol;

    Boolean      usePJL:1;
    UInt8        reserved0:7;
    UInt8        reserved1;
    UInt16       resolution;

    UInt8        reserved2[ 28 ];
} PrinterDescriptor;

/* If dataSize < 0, then StrLen( data ) is used instead */
/* Return true on success */
typedef Boolean (*ProtocolSendFunction)( UInt32 ref, void* data, Int32 dataSize );

/* This kind of function generates data for printer of type driver, sending with as
   many calls of SendFunction( sendFunctionRef, output, size ) as are needed.  If
   dataMatters is not set, then the content of the output does not matter--we're just
   trying to figure out the total data length (of course when calling SendFunction() with
   dataSize < 0, then the StrLen() must be right).  Returns true on success. */
/* The data must already include in itself any needed length and other information. */
typedef Boolean (*PrintDriverFunction)( DriverType driver, void* data,
    ProtocolSendFunction SendFunction, UInt32 sendFunctionRef, Boolean v );

/* The job number is for LPD/LPR and only the last three digits matter then, but it
   should be sequentially used.  The jobTitle is ignored for RAW, and in some LPD/LPR
   implementations. */
Boolean Print( PrinterDescriptor* printerP, PrintDriverFunction DriverFunction, void* data,
    Char* jobTitle, UInt32 jobNumber );
#endif
